/**
 * Title:        Galois Field<p>
 * Package:      ecc (Error-Correcting Codes)
 * Project:      EE/Ma 127b Error-Correcting Codes
 *               Project 1 - Reed-Solomon Decoder<p>
 * Copyright:    Copyright (c) Ling Li, 2001<p>
 * Company:      California Institute of Technology<p>
 * @author Ling Li
 * @email ling@cs.caltech.edu
 * @version 1.1
 */

/** TODO:
 *    1. test GF(p^m) for p != 2
 *    2. function ord(a), euclid(a,b)?, ...
 *    3. add/sub is awkward by using int as the number.
 *       Shall we change to array?
 *    4. How to share alpha_to/log_of between different instances?
 */

package ecc;

public class GF {
    /** GF(p^m), where p is a prime, n = p^m-1 */
    private int p, m, n;
    private int genpoly[];

    /** alpha_to: */
    private int alpha_to[];
    private int log_of[];

    public GF() {
        int poly[] = {1, 1};  // 1+x
        initialize(2, poly);  // GF(2)
    }

    public GF(int p, int genpoly[]) {
        // assert (p is a prime)
        initialize(p, genpoly);
    }

    public Object clone() {
        GF t = new GF();
        t.p = p; t.m = m; t.n = n;
        //t.genpoly = (int[]) genpoly.clone(); // genpoly is no longer needed
        t.alpha_to = (int[]) alpha_to.clone();
        t.log_of = (int[]) log_of.clone();
        return (Object) t;
    }

    public int power(int pow) { return alpha_to[(n + (pow % n)) % n];}
    public int log(int a) { return log_of[a];}
    public int max_num() { return n;}
    public int nOne(int n) { return (n % p); }

    private int[] toBits(int a) {
        //assert(a >= 0 && a <= n);
        int bit[] = new int[m];
        for (int i = 0; i < m; i++) {
            bit[i] = a % p;
            a /= p;
        }
        return bit;
    }

    private int toNum(int bits[]) {
        //assert(bits.length = m);
        int a = 0;
        for (int i = 0; i < m; i++) {
            a *= p; a += bits[i];
        }
        return a;
    }

    public int add(int a, int b) {
        //assert(a >= 0 && a <= n && b >=0 && b <= n)
        if (p == 2)
            return (a ^ b);
        else {
            int abits[], bbits[];
            abits = toBits(a);
            bbits = toBits(b);
            for (int i = 0; i < m; i++)
                abits[i] = (abits[i] + bbits[i]) % p;
            return toNum(abits);
        }
    }

    public int sub(int a, int b) {
        //assert(a >= 0 && a <= n && b >=0 && b <= n)
        if (p == 2)
            return (a ^ b);
        else {
            int abits[], bbits[];
            abits = toBits(a);
            bbits = toBits(b);
            for (int i = 0; i < m; i++)
                abits[i] = (p + abits[i] - bbits[i]) % p;
            return toNum(abits);
        }
    }

    public int mul(int a, int b) {
        //assert(a >= 0 && a <= n && b >=0 && b <= n)
        if (a == 0 || b == 0)
            return 0;
        return alpha_to[(log_of[a] + log_of[b]) % n];
    }

    public int inv(int a) {
        //assert(a != 0)
        if (a == 1)   // log_of[a] == 0
            return 1;
        return alpha_to[n - log_of[a]];
    }

    public int div(int a, int b) {
        //assert(b != 0)
        return mul(a, inv(b));
    }

    private void initialize(int _p, int []poly) {
        // assert (p is a prime)
        // assert(poly[poly.length-1] == 1???);
        p = _p; m = poly.length - 1;
        genpoly = (int []) poly.clone();

        /* n = power(p, m) - 1 */
        int i;
        if (m == 2)
            n = 1 << m;
        else
            for (n = 1, i = 0; i < m; i++) n *= p;
        n--;

        /** generate alpha_to and log_of */
        alpha_to = new int[n+1];
        log_of = new int[n+1];
        int mask = 1, alpha_m = 0;
        for (i = 0; i < m; i++) {
            alpha_to[i] = mask;
            alpha_m += mask * ((p - genpoly[i]) % p);
            mask *= p;
        }

        alpha_to[m] = alpha_m;

        if (p == 2) {
            for (i = m + 1; i < n; i++) {
                alpha_to[i] = alpha_to[i-1] << 1;
                if (alpha_to[i] >= mask)
                    alpha_to[i] = alpha_m ^ (alpha_to[i] ^ mask);
            }
        }
        else {
            for (i = m + 1; i < n; i++) {
                alpha_to[i] = alpha_to[i-1] * p;
                if (alpha_to[i] >= mask)
                    alpha_to[i] = add(alpha_m, alpha_to[i] % mask);
            }
        }
        alpha_to[n] = 0;    // we use n as the log of 0

        for (i = 0; i <= n; i++)
            log_of[alpha_to[i]] = i;
    }
}